/*
  Macro          JsonPP
  Author         Carlo Hogeveen
  Website        eCarlo.nl/tse
  Compatibility  TSE v4 upwards, all variants
  Version        v0.1.1   BETA   29 Aug 2024

  This quick & dirty tool takes the current block or buffer, copies it to a new
  buffer, and tries to "pretty print" it as JSON, reporting some JSON syntax
  errors.

  Pretty printing is a historical term for reformatting a text with a known
  syntax into a text with consistent indentation, line alignment, line breaks,
  etcetera.

  Syntax errors it will report:
    A different amount of { and [ versus } and ].
    Unmatched double quotes.

  Note: It can handle incorrect line-splits caused by TSE's line length limit.


  INSTALLATION

  Copy this file to TSE's "mac" directory and compile it there, for example by
  opening it there with TSE and applying the Macro -> Compile menu.

  Optionally add it to TSE's Potpourri menu.


  USAGE

  For example, open a JSON file in TSE and execute this macro.


  KNOWN ERROR
    The formatting might go wrong if both of these are true:
    - AND the file size is near or larger than 32Kb.
    - AND TSE uses the Unicode macro.
    Currently there is no reported case of this happening.
    I want to fix this, but that will not have high priority until a user
    requests it or I run into such a large .json file.


  HISTORY

  v0.0.0.1   UNPUBLISHED   15 Apr 2010
    Initial quick&dirty version.

  v0.1       BETA          28 Aug 2024
    Fixed a bug, enhanced the code a bit, and improved the documentation.

  v0.1.1     BETA          29 Aug 2024
    Documented a known error.

*/

integer indentation_increment = 2

proc Main()
  integer errors                   = 0
  integer indentations             = 0
  integer jsn_id                   = 0
  integer json_buffer_counter      = 0
  string  json_dpne [MAXSTRINGLEN] = SplitPath(CurrFilename(), _DRIVE_|_PATH_|_NAME_)
  integer line_from                = 0
  string  new_shift_text     [255] = ''
  string  old_shift_text     [255] = ''
  string  org_dpn   [MAXSTRINGLEN] = SplitPath(CurrFilename(), _DRIVE_|_PATH_|_NAME_)
  integer org_id                   = GetBufferId()
  integer pos_from                 = 0
  integer prev_line_len            = 0
  integer quote_problem            = FALSE
  integer tmp_id                   = 0
  string  token     [MAXSTRINGLEN] = ''
  integer token_counter            = 0
  integer token_from               = 0
  integer token_to                 = 0

  Message('JSON formatting and validation ... ')
  PushPosition()
  PushBlock()

  tmp_id = CreateTempBuffer()

  // Create a json buffer with a new free session name.
  repeat
    json_buffer_counter = json_buffer_counter + 1
    json_dpne = org_dpn + '_' + Str(json_buffer_counter) + '.json'
    if not FileExists(json_dpne)
      jsn_id = CreateBuffer(json_dpne)
      SetUndoOff()
    endif
  until jsn_id              >     0
     or json_buffer_counter > 10000

  // Copy the marked text or otherwise the whole original buffer to the temp
  // buffer, where we may break down the content.
  GotoBufferId(org_id)
  if not isBlockMarked()
  or not isBlockInCurrFile()
    MarkLine(1, NumLines())
  endif
  Copy()
  GotoBufferId(tmp_id)
  Paste()
  UnMarkBlock()

  // When too long lines are split, they may become separated by empty lines.
  while lFind('^$', 'gx')
    KillLine()
  endwhile

  // Concatenate the text to as few lines as possible.
  BegFile()
  prev_line_len = CurrLineLen()
  while Down()
    if prev_line_len + CurrLineLen() < MAXLINELEN
      Up()
      JoinLine()
    endif
    prev_line_len = CurrLineLen()
  endwhile

  // Convert the temp file to the future json file, which in this stage is
  // created to contain one token per line.
  BegFile()
  repeat
    token_counter = token_counter + 1
    case Chr(CurrChar())
      when ' ', Chr(9)
        token = ' '
      when '"'
        token_from    = CurrPos()
        token_to      = 0
        quote_problem = FALSE
        repeat
          if lFind('"', 'c+')
            if Chr(CurrChar(CurrPos() - 1)) <> '\'
              if (    CurrChar(CurrPos() + 1)  in _AT_EOL_, _BEYOND_EOL_)
              or (Chr(CurrChar(CurrPos() + 1)) in '{', '}', '[', ']', ':', ',', ' ')
                token_to = CurrPos()
              else
                quote_problem = TRUE
              endif
            endif
          else
            token_to = -1
          endif
        until token_to <> 0
        GotoPos(token_from)
        if token_to > 0
          token = GetText(token_from, token_to - token_from + 1)
        else
          token = '"'
        endif
        if token_to <= 0
        or quote_problem
          errors = errors + 1
          AddLine('!!! Error: Problem matching double quotes !!!', jsn_id)
        endif
      when '[', ']', '{', '}', ',', ':'
        token = Chr(CurrChar())
      otherwise
        // Should be a number.
        token_from = CurrPos()
        token_to   = 0
        if not lFind('["\[\]\{\},:]', 'cx+')
          EndLine()
        endif
        token_to = CurrPos() - 1
        token    = Trim(GetText(token_from, token_to - token_from + 1))
        GotoPos(token_from)
    endcase
    if token <> ' '
      // Message('Found token ', token_counter, ': ', token)
      AddLine(token, jsn_id)
    endif
    NextChar(Max(Length(token), 1))
    // Repair split lines.
    if  CurrPos()  >  255
    and CurrLine() <  NumLines()
      line_from = CurrLine()
      pos_from  = CurrPos()
      repeat
      until not Down()
      if CurrLine() > line_from
        old_shift_text = GetText(1, 255)
        MarkColumn(CurrLine(), 1, CurrLine(), 255)
        KillBlock()
        repeat
          Up()
          new_shift_text = GetText(1, 255)
          MarkColumn(CurrLine(), 1, CurrLine(), 255)
          KillBlock()
          EndLine()
          InsertText(old_shift_text, _INSERT_)
          old_shift_text = new_shift_text
        until CurrLine() == line_from
      endif
      GotoPos(pos_from - 255)
    endif
  until CurrLine() >= NumLines()
    and CurrPos()  >  CurrLineLen()

  // Concatenate and/or indent the tokens in the json file.
  GotoBufferId(jsn_id)
  token_counter = 0
  BegFile()
  repeat
    BegLine()
    token_counter = token_counter + 1
    // Message('Formatting token ', token_counter, ': ', GetText(1,CurrLineLen()))
    case GetText(1,CurrLineLen())
      when ':'
        Up()
        JoinLine()
        if Down()
          if (Chr(CurrChar()) in '{', '[', ']', '}', ',')
            Up()
          else
            Up()
            JoinLine()
          endif
        endif
      when ','
        Up()
        JoinLine()
      when '{', '['
        InsertText(Format('':(indentations * indentation_increment)), _INSERT_)
        indentations = indentations + 1
      when '}', ']'
        if indentations > 0
          indentations = indentations - 1
        else
          InsertLine('!!! Error: Decreasing indentation failed for the below token !!!')
          Down()
          errors = errors + 1
        endif
        InsertText(Format('':(indentations * indentation_increment)), _INSERT_)
      otherwise
        InsertText(Format('':(indentations * indentation_increment)), _INSERT_)
    endcase
  until not Down()

  if indentations <> 0
    errors = errors + 1
    AddLine('!!! Error: There is remaining indentation !!!')
  endif

  // Show result.
  PopPosition()
  PopBlock()
  if jsn_id > 0
    GotoBufferId(jsn_id)
    BegFile()
    SetUndoOn()
    UpdateDisplay()
    if errors == 0
      Message('No format errors   ')
      Delay(36)
    else
      if Query(Beep)
        Alarm()
      endif
      Find('!!! ?Error.*!!!', 'givx')
    endif
  else
    if Query(Beep)
      Alarm()
    endif
    Warn('ERROR: Could not create a new buffer.')
  endif

  AbandonFile(tmp_id)
  PurgeMacro(SplitPath(CurrMacroFilename(), _NAME_))
end Main

